#############################################################################
##
#W normalizer.gi               Package Coclass                   Bettina Eick
##

#############################################################################
##
#F QCentralizer( G ) . . . . . . . . . . . . . . . .centralizer in GL( d, Q )
##
## Returns a basis of the centralizer algebra C. Note that C forms a field
## and all elements of C \ {0} are in the desired centralizer
##
QCentralizer := function( G )
    local M, d, R, U, C;

    M := GroupByGenerators( G!.mats );
    d := DimensionOfMatrixGroup( M );
    R := MatrixAlgebra( Rationals, d );
    U := Subalgebra( R, GeneratorsOfGroup(M) );
    C := Centralizer( R, U );
    if not IsAbelian(C) then Print("warning: non-abelian \n"); fi;
    return BasisVectors( Basis( C ) );
end;

#############################################################################
##
#F ComplementsMod( G, M, N ) . . . . . .compute all complements to M/N in G/N
##
## This is a hack.
##
ComplementsMod := function( G, M, N )
    local nat, H, T, cl;
    nat := NaturalHomomorphismByNormalSubgroup( G, N );
    H := Image( nat );
    T := Image( nat, M );
    cl := Complementclasses( H, T );
    cl := Concatenation( List( cl, x -> Orbit( H, x ) ) );
    cl := List( cl, x -> PreImage( nat, x ) );
    return cl;
end;

#############################################################################
##
#F TryLifting( G, auto )
##
TryLifting := function( G, auto )
    local gens, imgs, gmat, imat, d, sys, c, r, m, n, i, j, k, sol, pcp; 

    # get info
    pcp := Pcp(G);
    gens := Igs(G);
    imgs := List( gens, x -> Image( auto, x ) );
    gmat := List( gens, x -> MappedVector( ExponentsByPcp(pcp,x), G!.mats ) );
    imat := List( imgs, x -> MappedVector( ExponentsByPcp(pcp,x), G!.mats ) );
    d := Length(G!.mats[1]);

    # set up system
    sys := NullMat( d^2, Length(gens)*d^2 );
    c := 0;
    for r in [1..Length(gens)] do
        m := imat[r];
        n := gmat[r];
        for i in [1..d] do
            for j in [1..d] do

                # add column c 
                c := c+1;
                for k in [1..d] do
                    sys[d*(i-1)+k][c] := sys[d*(i-1)+k][c] + m[k][j];
                    sys[d*(k-1)+j][c] := sys[d*(k-1)+j][c] - n[i][k];
                od;
            od;
        od;
    od;

    # solve
    sol := NullspaceMat( sys );
    sol := List( sol, x -> CutVector(x, d) );
    
    # check
    for i in [1..Length(sol)] do
        if RankMat(sol[i]) = d then 
            if ForAll([1..Length(gens)], x -> gmat[x]^sol[i]=imat[x]) then
                return sol[i];
            fi;
        fi;
    od;

    if Length(sol) > 0 then Error("check this!"); fi;
    return fail;
end;

#############################################################################
##
#W TranslateAutosToPcp( GG, G, autos )
##
TranslateAutosToPcp := function( GG, G, autos )
    local gens, trsl, auts, aut, imgs, naut;

    gens := Pcgs(GG);
    trsl := Pcp(G);
    auts := [];
    for aut in autos do
        imgs := List( gens, x -> Image( aut, x ) );
        imgs := List( imgs, x -> ExponentsOfPcElement( gens, x ) );
        imgs := List( imgs, x -> MappedVector( x, trsl ) );
        naut := GroupHomomorphismByImagesNC( G, G, AsList( trsl ), imgs );
        Add( auts, naut );
    od;
    return auts;
end;

#############################################################################
##
#W AutomorphismGroupPcp( G )
##
AutomorphismGroupPcp := function( G )
    local GG, A, H, auts, C;

    # compute aut group of GG
    GG := PcpGroupToPcGroup( G );
    IsPGroup(GG);
    A := AutomorphismGroupPGroup(GG);

    # translate to pc group
    H := PcGroupAutPGroup( A );

    # translate to G
    auts := TranslateAutosToPcp( GG, G, H!.autos );

    C := Group( auts, IdentityMapping(G) );
    SetSize( C, A.size );
    C!.pcgroup := H;
    return C;
end;

#############################################################################
##
#F QNormalizer( G ) . . . . . . . . . . . . . . . . . .normalizer in GL(d, Q)
##
## This function returns a record with entries centralizer, supplement and
## pcgroup. Supplement is a pc sequence for N_Q(G) mod C_Q(G) and pcgroup is
## the corresponding pc group.
##
QNormalizer := function( G )
    local c, A, a, B, b, l, U, u, i, N, M, L, cl, found, j, k, m;
    
    # get centralizer
    #Print("  compute centralizer \n");
    c := QCentralizer( G );

    # get automorphism group
    #Print("  compute automorphisms \n");
    A := AutomorphismGroupPcp( G );
    a := GeneratorsOfGroup( A );
    B := A!.pcgroup;
    b := Pcgs( B );
    l := Length( b );

    # add subgroup U of A to C
    U := TrivialSubgroup( B );
    u := [];
    for i in Reversed( [1..l] ) do
        #Print("  start lifting ",i,"\n");
        N := Subgroup( B, b{[i+1..l]} );
        M := Subgroup( B, b{[i..l]} );
        L := Normalizer( M, U );
        if not IsSubgroup( N, L ) then 
            N := Intersection( L, N );
            cl := ComplementsMod( L, N, U );
            found := false;
            j := 1;
            while not found and j <= Length(cl) do
                k := ModuloPcgs( cl[j], U )[1];
                k := ExponentsOfPcElement( b, k );
                k := MappedVector( k, a );
                m := TryLifting( G, k );
                if not IsBool(m) then 
                    Add( u, m );
                    U := cl[j];
                    found := true;
                fi;
                j := j + 1;
            od;
        fi;
    od; 

    return rec( centralizer := c, supplement := u, autgrp := U );
end;


#############################################################################
##
#F CheckUniserials(s,p)
##
## Prints info (for all uniserials): 
##   cent = dimension of QCentralizer
##   know = dimension of cyclotomic field generated by the center
##                  (know should be at most cent)
##  -- 
##   norm = order of QNormalizer supplement
##   know = order of normalizer in W(s) 
##                  (know should be norm or norm/2)
##
CheckUniserials := function(s,p)
    local N, uni, i, U, L, norm;

    # get uniserials
    N := WreathFiniteNormalizer(s,p);
    uni := UniserialSubgroups(N, false);

    # loop
    for i in [1..Length(uni)] do
        U := SubgroupByIgs(N, uni[i].repr);
        L := SubgroupByIgs(N, uni[i].norm);

        # compute normalizer
        U!.mats := List(Igs(U), x -> MappedVector(Exponents(x), N!.mats));
        norm := QNormalizer(U);

        # print info
        if Length(norm.centralizer) <> Size(Center(U))/2 then
            Print(i," is special \n"); 
        else
            Print(i," is ordinary \n"); 
        fi;
        Print("      cent: ",Length(norm.centralizer));
        Print("  know: ",Size(Center(U))/2);

        Print("   -- norm: ",Size(norm.autgrp));
        Print("  know: ",Index(L, Centralizer(L,U)));
        Print("\n");
    od;
end;

