#############################################################################
##
#W  actions.gi                  Package Coclass                  Bettina Eick
##

#############################################################################
##
#F OperModLat( base, lat, mat, f )
##
ReduceModLat := function( base, lat, p )
    base := LatticeIntersection( p * base, p * lat );
    return 1/p * base;
end;

#############################################################################
##
#F CoefficientsMod( vec, base, lat, p )
##
CoefficientsMod := function( vec, base, lat, p )
    local full, solv;
    full := Concatenation( base, lat );
    solv := SolutionIntMat( p*full, p*vec );
    if IsBool(solv) then return fail; fi;
    return solv{[1..Length(base)]};
end;

#############################################################################
##
#F SolutionDiagMat( mat, vec )
##
SolutionDiagMat := function( mat, vec )
    local new, i;
    new := List( mat, x -> 0 );
    for i in [1..Length(vec)] do
        new[i] := vec[i] / mat[i][i];
    od;
    return new;
end;

#############################################################################
##
#F OpersModLat( base, lat, mat, p )
##
OpersModLat := function( base, lat, mats, p )
    local full, norm, N, Q, P, d, acts, mat, imgs, coef, f;

    full := p * Concatenation( base, lat );
    norm := NormalFormIntMat( full, 13 );
    N := norm.normal;
    Q := norm.rowtrans;
    P := norm.coltrans;
    d := Length( base );
    f := GF(p);

    acts := [];
    for mat in mats do
        imgs := p * (base * mat);
        coef := List( imgs, x -> SolutionDiagMat( N, x*P ) * Q );
        coef := List( coef, x -> x{[1..d]} );
        coef := Immutable( coef * One( f ) );
        ConvertToMatrixRep( coef, f );
        Add( acts, coef );
    od;
    return acts;
end;

#############################################################################
##
#F NullspaceMod( mat, lat, p )
##
NullspaceMod := function( mat, lat, p )
    local d, a, b, l, f;
  
    # get nullspace
    d := Length(mat);
    a := Concatenation(lat, mat);
    b := TriangulizedNullspaceMat( a );
    b := b{[1..d]}{[d+1..2*d]};

    # reduce to basis
    l := p * b;
    l := NormalFormIntMat( l, 2 ).normal;
    f := 1/p * AdditiveFactorPcp( l, p*lat, 0 ).prei;

    # normalize
    return f;
end;

#############################################################################
##
#F ActionRecord( N )
##
ActionRecord := function( N )
    local p, s, q, d, I, G, C, c, g, A, v, w, m, i, l;

    # set up
    p := N!.prime;
    s := N!.param;
    q := p^(s-1);
    d := q*(p-1);
    I := IdentityMat(d);

    # get groups and special elements
    G := N!.pointgroup;
    C := Centre( G );
    c := Igs(C)[1];
    g := Pcp( N, C );
    A := rec( gens := g );
    A.nath := NaturalHomomorphismByPcp( g );

    # get some special matrices
    v := PcpToMat( N, N!.mats, c ) - I; 
    w := PcpToMat( N, N!.mats, Product(G!.mins) ) - I;
    m := List( g, x -> PcpToMat( N, N!.mats, x ) );

    # compute p^i lattice bases and their inverses
    Print("  lattices \n");
    A.lats := [I];
    for i in [2..q] do
        A.lats[i] := NormalFormIntMat( A.lats[i-1] * w, 2 ).normal;
    od;

    # inverses
    Print("  inverses \n");
    A.invs := List( A.lats, x -> x^-1 );

    # compute F = Fix_C(V/T) for each lattice T
    Print("  fixedpoints \n");
    A.fixs := List( [1..q], x -> NullspaceMod(v, A.lats[x], p));

    # determine action on fixed points
    Print("  action \n");
    A.acts := List( [1..q], x -> OpersModLat(A.fixs[x], A.lats[x], m, p));

    # add invariant series
    Print("  invariant series \n");
    A.sers := [];
    for i in [1..q] do
        l := List( G!.mins, x -> PcpToMat( g, A.acts[i], x ) );
        A.sers[i] := InvariantFlag( l, q, GF(p) );
    od; 

    # initialize fixedpoint-bases list
    A.fbas := List( A.sers, x -> [] );
    A.prime := p;

    # that's it
    return A;
end;

#############################################################################
##
#F AddFixedPointsBasis( A, l, r )
##
AddFixedPointBasis := function( A, l, r )
    local p, m, base, subb, spcs, i, v;

    # check
    if IsBound( A!.fbas[l][r] ) then return A!.fbas[l][r]; fi;

    # set up
    p := A!.prime;
    m := IdentityMat( r, GF(p) );

    # get base
    base := List( A!.sers[l], x -> KroneckerProduct( m, [x] ) );

    # loop
    subb := [];
    spcs := [];
    for i in [1..Length(base)] do
        for v in base[i] do 
            AddVectorEchelonBase( subb, v ); 
        od;
        spcs[i] := ShallowCopy( subb );
    od;

    A!.fbas[l][r] := Reversed(spcs);
    return A!.fbas[l][r];
end;

#############################################################################
##
#F  AddActions( A, Ni, Ci )
##
AddActions := function( A, Ni, Ci )
    local Gi, N, s, p, q, c, j, i;

    # set up
    Gi := Ni!.pointgroup;
    N := Parent(Ni);
    s := N!.param;
    p := N!.prime;
    i := Ni!.index;
    q := p^(s-1);

    # add actions of Ci
    if i = 0 then
        A.cent := List( [1..q], x -> [] );
    else
        c := Ci!.mats;
        A.cent := List( [1..q], x -> OpersModLat(A.fixs[x], A.lats[x], c, p));
    fi;

    # determine conjugates of N if necessary
    if not COUNTONLY and TRANSLATE and not IsBound(A.conj) then
        A.conj := [];
        for j in [1..p^(s-1)] do
            A.conj[j] := List( N!.mats, x -> A.lats[j] * x * A.invs[j] );
        od;
    fi;

end;

#############################################################################
##
#F ActionRecordByConjugaction( N )
##
ActionRecordByConjugaction := function( N )
    local p, s, q, d, I, G, C, c, g, A, v, w, m, n, i, l;

    # set up
    p := N!.prime;
    s := N!.param;
    q := p^(s-1);
    d := q*(p-1);
    I := IdentityMat(d);

    # get group and natural action
    G := N!.pointgroup;
    C := Centre( G );
    c := Igs(C)[1];
    g := Pcp( N, C );
    w := PcpToMat( N, N!.mats, Product(G!.mins) ) - I;

    # set up
    A := rec( gens := g );
    A.nath := NaturalHomomorphismByPcp( g );
    A.lats := [I];
    A.invs := [I];
    A.fixs := [];
    A.acts := [];
    A.sers := [];

    # lattices and inverses
    for i in [2..q] do
        A.lats[i] := NormalFormIntMat( A.lats[i-1] * w, 2 ).normal;
        A.invs[i] := A.lats[i]^-1;
    od;

    # actions
    for i in [1..q] do

        # conjugate action
        n := List( N!.mats, x -> A.lats[i]*x*A.invs[i] );
        m := List( g, x -> PcpToMat( N, n, x ) );
        v := PcpToMat( N, n, c ) - I; 

        A.fixs[i] := NullspaceMod(v, A.lats[1], p);
        A.acts[i] := OpersModLat(A.fixs[i], A.lats[1], m, p);

        l := List( G!.mins, x -> PcpToMat( g, A.acts[i], x ) );
        A.sers[i] := InvariantFlag( l, q, GF(p) );
    od; 

    # initialize fixedpoint-bases list
    A.fbas := List( A.sers, x -> [] );
    A.prime := p;

    # that's it
    return A;
end;

