#############################################################################
##
#W  cohomgen.gi                 Package Coclass                  Bettina Eick
##

#############################################################################
##
#F AddToSystemCC( sys, t )
##
AddToSystemCC := function( sys, t )
    local i, j, v;

    # transpose 
    for i in [1..sys.len] do
        if IsBound( t[i] ) then t[i] := TransposedMat(t[i]); fi;
    od;

    # add
    for j in [1..sys.dim] do
        v := [];
        for i in [1..sys.len] do
            if IsBound( t[i] ) then
                v[i] := t[i][j];
            else
                v[i] := sys.zero;
            fi;
        od;
        AddSystemFFE( sys, Concatenation(v) );
    od;
end;

#############################################################################
##
#F SiftIntoIgs( igs, dep, wrd, i, j, p )
##
SiftIntoIgs := function( igs, dep, wrd, i, j, p )
    local g, h, k, e;

    if i <> j then 
        g := igs[j]^igs[i];
        h := CompactedWord( ConjugatedWord( wrd[j], wrd[i] ) );
        k := Position( dep, Depth( g ) );
    else
        g := igs[i]^p;
        h := CompactedWord( PoweredWord( wrd[i], p ) );
        k := Position( dep, Depth( g ) );
    fi;

    # if the word is too long, then try another one
    if Length(h) > 5000 then return true; fi;

    while not IsBool( k ) and not IsBool( igs[k] ) do

        # if the word is very short, then just insert
        if Length( h ) < Length( wrd[k] ) then 
            igs[k] := g;
            wrd[k] := h;
            return true;
        fi;
            
        # otherwise divide off
        e := ReducingCoefficient( g, igs[k] );
        g := g * igs[k]^-e;
        h := CompactedWord( Concatenation( h, PoweredWord( wrd[k], -e ) ) );
        k := Position( dep, Depth( g ) );
    od;

    # finally just add g and h if possible
    if not IsBool( k ) then 
        igs[k] := g;
        wrd[k] := CompactedWord( h );
        return true;
    fi;

    return false;
end;
    
#############################################################################
##
#F AddRewritingMG( C )
##
AddRewritingMG := function( C )
    local deps, gens, igs, wrd, i, d, j, flg, tmp;
 
    # set up
    deps := List( C.facigs, Depth );
    gens := C.factor;
    igs := List( deps, x -> fail );
    wrd := List( deps, x -> fail );

    # add gens
    for i in [1..Length(gens)] do
        d := Depth(gens[i]);
        j := Position( deps, d );
        igs[j] := gens[i];
        wrd[j] := [[i,1]];
    od;

    # add conjugates
    while ForAny(igs, IsBool) do
        flg := false;
        for i in [1..Length(deps)] do
            if not IsBool( igs[i] ) then
                for j in [i..Length(deps)] do
                    if not IsBool( igs[j] ) then  
                        tmp := SiftIntoIgs( igs, deps, wrd, i, j, C.char );
                        if tmp then flg := true; fi;
                    fi;
                od;
            fi;
        od;
        if not flg then Error("cannot find expressions"); fi;
    od;
    C.facigs := igs;
    C.facwrd := wrd;
end;

#############################################################################
##
#F WordInGensMG( C, g )
##
WordInGensMG := function( C, g )
    local w, i;
    w := WordOfVectorCR( ExponentsByIgs( C.facigs, g ) );
    for i in [1..Length(w)] do
        w[i] := Concatenation(List( [1..w[i][2]], x -> C.facwrd[w[i][1]] ));
    od;
    return CompactedWord( Concatenation( w ) );
end;

#############################################################################
##
#F AddRelatorsMG( C, UU )
##
AddRelatorsMG := function( C, UU )
    local iso1, F1, iso2, F2, f2, f1, uu, r, i;
 
    iso1 := IsomorphismFpGroup(UU);
    F1 := Range( iso1 );
    iso2 := IsomorphismSimplifiedFpGroup(F1);
    F2 := Range( iso2 );

    # get generators
    f2 := GeneratorsOfGroup(F2);
    f1 := List( f2, x -> PreImagesRepresentative( iso2, x ) );
    uu := List( f1, x -> PreImagesRepresentative (iso1, x ) );
    C.factor := uu;

    # get relations
    r := RelatorsOfFpGroup(F2);
    r := List( r, ExtRepOfObj );
    for i in [1..Length(r)] do
        r[i] := List( [1..Length(r[i])/2], j -> [r[i][2*j-1], r[i][2*j]] );
    od;
    C.rels := r;
end;

#############################################################################
##
#F AddActionMG( C, A, l )
##
AddActionMG := function( C, A, l )
    local acts, mats;

    # get action corresponding to factor and super
    acts := List( C.factor, Exponents );
    acts := List( acts, x -> MyMappedVector( x, A!.acts[l], false ) );
    mats := List( C.super, Exponents );
    mats := List( mats, x -> MyMappedVector( x, A!.acts[l], false ) );

    # add info
    C.mats := acts;
    C.smats := Concatenation( mats, A.cent[l] );
    C.one := acts[1]^0;
    AddInversesCR( C );
end;

#############################################################################
##
#F CohomologyRecordMG( N, A, U, H )
##
CohomologyRecordMG := function( N, A, U, H )
    local p, s, q, UU, HH, C;

    # set up
    p := N!.prime;
    s := N!.param;
    q := p^(s-1);
    UU := Image( A.nath, U );
    HH := Image( A.nath, H );

    # set up record
    C := rec( facigs := Cgs(UU),
              super := Pcp( HH, UU ),
              dim := q, char := p, field := GF(p) );

    # rewrite
    #Print("  add relators \n");
    AddRelatorsMG( C, UU );
    #Print("  add rewriting \n");
    AddRewritingMG( C );
    
    # add enlarged igs of U
    C.repigs := Cgs(U);
    C.repwrd := List( Cgs(U), x -> WordInGensMG( C, Image( A.nath, x ) ) );
    return C;
end;

#############################################################################
##
#F OneCc( C )
##
OneCc := function( C )
    local sys, i, w, tran, null;

    # create system
    sys := CRSystem( C.dim, Length(C.mats), C.char ); 
    sys.deps := []; sys.blen := 0; 

    # loop over relators
    for i in [1..Length(C.rels)] do
        w := CollectedOneCR( C, C.rels[i] );
        AddToSystemCC( sys, w );
        #Print("    collected relator ",i," system has length ",sys.blen,
        #      " in ",sys.len*sys.dim,"\n");
    od;

    if Length( sys.base ) = 0 then
        return IdentityMat( sys.dim * sys.len ) * One(C.field);
    fi;

    # solve system
    tran := TransposedMat( sys.base );
    null := TriangulizedNullspaceMat( tran );
    return null;
end;

#############################################################################
##
#F OneCb( C )
##
OneCb := function( C )
    local mat, i, v;
    mat := [];
    for i in [1..C.dim] do
        v := List( [1..Length(C.mats)], j -> C.mats[j][i] - C.one[i] );
        Add( mat, Flat( v ) );
    od;
    return SemiEchelonMat( mat ).vectors;
end;

#############################################################################
##
#F InvCohMap( coh, base )
##
InvCohMap := function ( coh, base )
    local mat, new, dep, i, inv;
    mat := MutableIdentityMat( Length(base[1]), coh.fld );
    dep := List( base, PositionNonZero );
    new := List( base, function ( x ) return MutableCopyMat( x ); end );
    for i  in [ 1 .. Length(base[1]) ]  do
        if not i in dep  then Add( new, mat[i] ); fi;
    od;
    inv := new ^ -1;
    inv := inv{[1..Length(base[1])]}{[1..Length(base)]};
    inv := Immutable( inv );
    ConvertToMatrixRep( inv, coh.fld );
    return inv;
end;

#############################################################################
##
#F OneCohom( C )
##
OneCohom := function( C )
    local cc, cb, cc2, cb2, coh;

    # compute cocycles and cobounds
    cb := OneCb( C ); 
    cc := OneCc( C );

    # set up cohomology record
    coh := rec( gcc := cc, gcb := cb );
    coh.fld := GF( C.char ); 

    # add decription of the factor gcc/gcb
    coh.factor := AdditiveFactorPcp( coh.gcc, coh.gcb, C.char );

    # compute linear mapping extend coh.gcc to an full basis
    coh.invgcc := InvCohMap( coh, coh.gcc );

    # compute linear mapping to factor
    coh.invfac := Immutable( coh.invgcc * coh.factor.imgs );
    ConvertToMatrixRep( coh.invfac, C.field );

    # return
    return coh;
end;

#############################################################################
##
#F Conversion functions
##
CocToCCElement := function( coh, coc )
    if Length( coh.gcc ) = 0 then return []; fi;
    return coc * coh.invgcc;
end;

CocToFactor := function( coh, coc )
    return coc * coh.invfac;
end;

#############################################################################
##
#F EvaluateCocycleMG( C, coc, w )
##
EvaluateCocycleMG := function( C, coc, w )
    local  n, m, i, e, j, k;
    n := 0 * coc[1];
    m := C.one;
    for i  in Reversed( [ 1 .. Length(w) ] )  do
        j := w[i][1];
        e := w[i][2];
        if e > 0  then
            for k  in [ 1 .. e ]  do
                n := n + coc[j] * m;
                m := C.mats[j] * m;
            od;
        elif e < 0  then
            for k  in [ 1 .. - e ]  do
                m := C.invs[j] * m;
                n := n - coc[j] * m;
            od;
        fi;
    od;
    return n;
end;

#############################################################################
##
#F OneCohomologyOperation( C, cc )
##
OneCohomologyOperation := function( C, cc )
    local lin, i, g, m, s, j, coc, img;

    # set up
    lin := List( C.smats, x -> [] );

    # get the upper action
    for i in [1..Length( C.super )] do
        g := C.super[i]^-1;
        m := C.smats[i];
        s := List( C.factor, x -> WordInGensMG( C, x^g ) );
        for j in [1..Length( cc.factor.prei )] do
            coc := CutVector( cc.factor.prei[j], Length( C.factor ) );
            img := List( s, x -> EvaluateCocycleMG( C, coc, x ));
            img := List( img, x -> x * m );
            lin[i][j] := CocToFactor( cc, Flat( img ) );
        od;
        lin[i] := Immutable( lin[i] );
        ConvertToMatrixRep( lin[i], C.field );
    od;

    # get the centralizer
    for i in [Length( C.super )+1 .. Length(C.smats) ] do
        m := C.smats[i];
        s := List( C.factor, x -> WordInGensMG( C, x ) );
        for j in [1..Length( cc.factor.prei )] do
            coc := CutVector( cc.factor.prei[j], Length( C.factor ) );
            img := List( s, x -> EvaluateCocycleMG( C, coc, x ));
            img := List( img, x -> x * m );
            lin[i][j] := CocToFactor( cc, Flat( img ) );
        od;
        lin[i] := Immutable( lin[i] );
        ConvertToMatrixRep( lin[i], C.field );
    od;
 
    return lin;
end;

