#############################################################################
##
#W  general.gi                  Package Coclass                  Bettina Eick
##

#############################################################################
##
#F DepthVector(vec)
##
DepthVector := function(vec)
    local i;
    for i in [1..Length(vec)] do
        if vec[i] <> 0*vec[i] then return i; fi;
    od;
    return Length(vec)+1;
end;

#############################################################################
##
#F MyMappedVector( e, m, s ) . . . . . . . . . . . . . . . .includes inverses
##
MyMappedVector := function( arg )
    local e, m, s, r, i;
    r := false;
    e := arg[1];
    m := arg[2];
    if Length(arg) = 2 then s := []; else s := arg[3]; fi;

    for i in [1..Length(e)] do
        if e[i] > 0 then
            if IsBool(r) then
                r := m[i]^e[i];
            else
                r := r * m[i]^e[i];
            fi;
        elif e[i] < 0 then
            if not IsBound( s[i] ) then s[i] := m[i]^-1; fi;
            if IsBool(r) then
                r := s[i]^-e[i];
            else
                r := r * s[i]^-e[i];
            fi;
        fi;
    od;
    if IsBool(r) then r := m[1]^0; fi;
    return r;
end;

#############################################################################
##
#F CanonicalBaseMat(mat, F)
##
CanonicalBaseMat := function( mat, F )
    local i, m, n, hds, mhd, don, j, h, new;

    # set up and catch trivial problems
    if Length( mat ) = 0 then return mat; fi;
    if not IsMutable( mat ) then mat := MutableCopyMat( mat ); fi;
    for i in [1..Length(mat)] do
        mat[i] := ShallowCopy( mat[i] );
        ConvertToVectorRep( mat[i], F );
    od;

    # get info
    m := Length(mat);
    n := Length(mat[1]);
    hds := List( mat, x -> DepthVector(x) );
    mhd := Minimum( hds );
    don := List( mat, x -> false );

    while mhd <= n do
        j := Position( hds, mhd );
        h := hds[j];
        don[j] := true;

        # reset leading coefficient
        if mat[j][h] <> One(F) then
            MultRowVector( mat[j], mat[j][h]^-1 );
        fi;

        # clear entries in mat
        for i in [1..Length(mat)] do
            if i <> j and mat[i][h] <> Zero(F) then
                AddRowVector( mat[i], mat[j], -mat[i][h] );
                if not don[i] then
                    hds[i] := DepthVector( mat[i] );
                fi;
            fi;
        od;

        # reset mhd
        mhd := n+1;
        for i in [1..Length(mat)] do
            if not don[i] then
                mhd := Minimum( mhd, hds[i] );
            fi;
        od;
    od;

    new := List([1..Length(mat)], x -> false );
    for i in [1..Length(mat)] do
        if hds[i] <= n then
            new[hds[i]] := mat[i];
        fi;
    od;
    new := Filtered( new, x -> not IsBool( x ) );

    new := Immutable(new);
    ConvertToMatrixRep( new, F );
    return new;
end;

#############################################################################
##
#F InvariantFlag( mats, d, F )
##
InvariantFlag := function( mats, d, F )
    local p, M, I, V, t, i, j, flag, zero; 

    # set up 
    p := Characteristic(F);
    mats := Filtered( mats, x -> IsInt( Order(x) / p ) );
    if Length(mats) = 0 or d = 1 then return IdentityMat( d, F ); fi;

    # set up and catch trivial cases
    I := IdentityMat( d, F );
    M := List( mats, x -> x - I ); 

    # compute invariant subspaces
    V := [ I ];
    zero := 0 * I[1];
    t := Length( M );
    i := 1;
    while Length( V[i] ) > 0 do
        if i > d+1 then Error( "M[i] are not nilpotent" ); fi;
        i := i + 1;
        V[i] := Concatenation( List( [1..t], x -> V[i-1] * M[x] ) );
        V[i] := CanonicalBaseMat( V[i], F );
    od;

    # compute flag
    flag := [];
    t := Length( V );
    for i in [ t, t-1 .. 2 ] do
        Append( flag, SemiEchelonFactorBase( V[i-1], V[i] ) );
    od;
    return flag;
end;

#############################################################################
##
#F CheckMatrixRepresentationByPcp(g, m)
##
CheckMatrixRepresentationByPcp := function( g, m )
    local r, i, e, s, t, j;

    r := RelativeOrdersOfPcp( g );
    for i in [1..Length(g)] do

        if r[i] > 0 then
            e := ExponentsByPcp( g, g[i]^r[i] );
            s := m[i]^r[i];
            t := MappedVector( e, m );
            if s <> t then Error(i," is wrong power"); fi;
        fi;

        for j in [1..i-1] do
            e := ExponentsByPcp( g, g[i]^g[j] );
            s := m[i]^m[j];
            t := MappedVector( e, m );
            if s <> t then Error(i,j, " is wrong conjugate"); fi;

        od;
    od;
end;

#############################################################################
##
#F CheckPointGroupRepresentation(G)
##
CheckPointGroupRepresentation := function( G )
    local g, m, i;

    # get elements and mats
    g := Pcp( Parent(G) );
    m := StructuralCopy( G!.mats );

    # check and kill cyclotomics
    for i in [1..Length(g)] do
        if RelativeOrderPcp( g[i] ) = 0 then
            if not ForAll( m{[i+1..Length(m)]}, x -> m[i]*x = x*m[i] ) then
                Error("wrong centralizer");
            fi;
            m[i] := m[i]^0;
        fi;
    od;

    CheckMatrixRepresentationByPcp( g, m );
end;

#############################################################################
##
#F NrOrbits( mats, subs )
##
NrOrbits := function( mats, subs )
    local m, M, nat, L, cl, n, c, l, g, f, p;

    # check that the subspace is non-trivial
    if Length(subs) = 0 then return 1; fi;
    p := Characteristic(Field(subs[1][1]));

    # check that the action is non-trivial
    if Length(mats) = 0 then return p^Length(subs); fi;

    # get mat group
    m := List( mats, x -> MTX.InducedActionSubMatrix( x, subs ) );
    M := Group( m );

    # get image and classes
    nat := IsomorphismPcGroup( M );
    L := Image( nat );
    cl := ConjugacyClasses(L);

    # count
    n := 0;
    for c in cl do
        l := Size(c);
        g := Representative(c);
        g := PreImagesRepresentative( nat, g );
        f := TriangulizedNullspaceMat( g - g^0 );
        n := n + l * p^Length(f);
    od;

    # return number of orbits
    return n / Size(L);
end;

#############################################################################
##
#F ElementsOutside( subb, full )
##
ElementsOutside := function( subb, full )
    local n, m, f, V, elmV, fact, U, elmU, elmF, i;

    # the trivial case
    if Length(full) = Length(subb) then return []; fi;

    # catch arguments
    n := Length( full ) - Length( subb );
    m := Length( subb );
    f := Field( full[1][1] );

    # the elements of the factor
    V := f^n;
    fact := BaseSteinitzVectors( full, subb ).factorspace;
    elmV := Filtered( Elements(V), x -> x <> 0 * x );
    elmV := List( elmV, x -> x * fact );

    # if the subspace is trivial, then this is it
    if Length( subb ) = 0 then return elmV; fi;

    # otherwise we need to add subspace elements
    U := f^m;
    elmU := Elements(U);
    elmU := List( elmU, x -> x * subb );
    elmF := Concatenation( List( elmV, x -> x + elmU ) );
    return elmF;
end;


