#############################################################################
##
#W pointgrps.gi                Package Coclass                   Bettina Eick
##
## Create group G given as a pcp group with entries
##   G!.param = s
##   G!.prime = p
##   G!.mins = minimal generators of G mod G'
##   G!.maxi = maximal subgroups of G
##   G!.mats = matrix action for the igs of G
##

#############################################################################
##
#F MatrixPointGroup( s, p )
##
MatrixPointGroup := function( s, p )
    local x, f, m, h, i, I, J, t, j, l, n, M;

    # get first generator
    x := Indeterminate(Rationals);
    f := Sum( List( [1..p], i -> x^(i-1) ) );
    m := [CompanionMat(f)];
    h := [2..p]; Add( h, 1); h := PermList( h );

    # now start computing wreath products
    for i in [1..s-1] do
        I := m[1]^0;
        J := List( [1..p], x -> I );
        t := [KroneckerProduct( PermutationMat( h, p ), I )];
        for j in [1..p] do
            for l in m do
                n := ShallowCopy( J );
                n[j] := l;
                n := DiagonalBlockMat( n );
                Add( t, n );
            od;
        od;
        m := ShallowCopy( t );
    od;

    # get matrix group
    M := Group( m, m[1]^0 );
    M!.prime := p;
    M!.param := s;

    return M;
end;

#############################################################################
##
#F PcpPointGroup( s, p )
##
PcpPointGroup := function( s, p )
    local C, H, c, act, G, g, h, j, i, d, max, M;

    # start with cyclic group
    C := CyclicGroup( IsPermGroup, p );
    c := GeneratorsOfGroup(C)[1];
    H := AbelianPcpGroup( 1, [p] );
    act := GroupHomomorphismByImagesNC( H, C, Igs(H), [c] );

    # wreath it
    G := H; for i in [1..s-1] do G := WreathProduct( G, H, act ); od;
    SetIsNilpotentGroup( G, true );

    # add info
    G!.param := s;
    G!.prime := p;

    # add special generators mod G'
    g := Reversed( Igs(G) );
    h := [];
    j := 0;
    for i in [1..s] do
        j := p*j + 1;
        Add( h, g[j] );
    od; 
    h := Reversed(h);
    G!.mins := h;

    # add maximal subgroups
    d := Cgs( DerivedSubgroup( G ) );
    max := List(h, x->Subgroup(G, Concatenation(d, Difference(h,[x]))));
    G!.maxi := max;

    # add matrix rep
    M := MatrixPointGroup(s,p);
    G!.mats := GeneratorsOfGroup(M);

    # that's it
    return G;
end;

#############################################################################
##
#F WreathPointGroup( G, s, i, p )
##
InstallMethod( WreathPointGroup, true, [IsGroup,IsInt], 0, 
function( G, i )
    local s, p, g, U, C;

    # take params and take generator
    s := G!.param;
    p := G!.prime;
    g := Product(G!.mins)^(p^i);

    # get subgroup 
    if i = s-1 then 
        U := G; 
    elif i = 0 then 
        U := Subgroup(Parent(G),[g]);
    else
        U := Centralizer(G, g);
        U := SubgroupByIgs(Parent(G), Cgs(U));
    fi;

    # add infos
    U!.param := s;
    U!.prime := p;
    U!.index := i;

    # add centre
    C := Subgroup(Parent(G), [g]);
    SetCentre(U, C);

    return U;
end );

InstallOtherMethod( WreathPointGroup, true, [IsInt,IsInt,IsInt], 0, 
function( s, i, p )
    return WreathPointGroup(PcpPointGroup(s,p),i);
end );
    
InstallOtherMethod( WreathPointGroup, true, [IsInt,IsInt], 0, 
function( s, p )
    return WreathPointGroup(PcpPointGroup(s,p), s-1); 
end );

