#############################################################################
##
#W  uniserial.gi               Package Coclass                   Bettina Eick 
##
##  Compute the uniserial subgroups of a maximal point group G up to 
##  conjugacy in its normalizer. 
##
##  Also, enable the restriction to covered uniserial groups only.
##

#############################################################################
##
#F ConsiderClass( H, cls, pcp, p )
##
ConsiderClass := function( H, cls, pcp, p )
    local U, N, C, M, tmp, i;

    # set up class record
    U := SubgroupByIgs( H, cls.repr );
    N := SubgroupByIgs( H, cls.norm );
    
    C := rec( );
    C.group  := U;
    C.super  := Pcp( N, U );
    C.factor := Pcp( U, GroupOfPcp( pcp ) );
    C.normal := pcp;

    # add extension info
    AddFieldCR( C );
    AddRelatorsCR( C );
    AddOperationCR( C );
    AddInversesCR( C );

    # compute complements
    tmp := ComplementClassesCR( C );

    # add info
    for i in [1..Length(tmp)] do
        tmp[i].rump := RumpDepth( tmp[i].repr, p );
        tmp[i].repr := Igs( tmp[i].repr );
        tmp[i].norm := Igs( tmp[i].norm );
        tmp[i].unis := cls.unis;
    od;
    return tmp;
end;

#############################################################################
##
#F ExternalSpaces( sub, n, p )
##
ExternalSpaces := function( sub, n, p )
    local V, all, U, C, cov, uni, i;

    # set up
    V := GF(p)^n;
    sub := List( sub, x -> Subspace( V, x ) );
    all := AllSubspaces(n, p);
    all := List( all, x -> Subspace( V, x ) );
    all := Concatenation( [V], all );

    # filter non-uniserial ones
    for U in sub do
        all := Filtered( all, x -> not IsSubspace(U,x) );
    od;

    # find uniserials
    C := Union( List( sub, Elements ) );

    # split up
    uni := [];
    cov := [];
    for i in [1..Length(all)] do
        if IsSubset( C, all[i] ) then 
            Add( cov, BasisVectors(Basis(all[i])) ); 
        else
            Add( uni, BasisVectors(Basis(all[i])) );
        fi;
        all[i] := false;
    od;

    if COVEREDGP then
        return [[], cov];
    else
        return [uni, cov];
    fi;
end;

#############################################################################
##
#F SubgroupClassesFirstStep( N, G, D, sub )
##
SubgroupClassesFirstStep := function( N, G, D, sub )
    local pcp, n, p, spc, all, mat, i, rep, tmp, fac, uni;

    # set up layer
    pcp := Pcp(G,D);
    n := Length(pcp);
    p := N!.prime;

    # action and subspaces
    spc := List( sub, x -> List( Igs(x), y -> ExponentsByPcp(pcp,y)));
    spc := List( spc, x -> x * One(GF(p)) );
    all := ExternalSpaces( spc, n, p );

    # set up for orbits
    fac := Pcp(N);
    mat := LinearActionOnPcp( fac, pcp );

    # the uniserials first 
    uni := PcpOrbitsStabilizers( all[1], fac, mat, OnBasesCase );
    for i in [1..Length( uni )] do
        rep := SubgroupBySubspace( pcp, uni[i].repr );
        tmp := rec();
        tmp.repr := Igs( rep );
        tmp.norm := uni[i].stab;
        tmp.rump := RumpDepth( rep , p );
        tmp.unis := true;
        uni[i] := tmp;
    od;

    # the covereds seconds
    all := PcpOrbitsStabilizers( all[2], fac, mat, OnBasesCase );
    for i in [1..Length( all )] do
        rep := SubgroupBySubspace( pcp, all[i].repr );
        tmp := rec();
        tmp.repr := Igs( rep );
        tmp.norm := all[i].stab;
        tmp.rump := RumpDepth( rep , p );
        tmp.unis := false;
        all[i] := tmp;
    od;

    return Concatenation( uni, all );
end;

#############################################################################
##
#F CheckCoclass( G, U, cc, d, deps )
##
CheckCoclass := function( G, U, cc, d, deps )
    local L, c;

    # easy check
    if (Length(U.repr) - d) > cc then return false; fi;

    # hard check
    L := SubgroupByIgs( G, U.repr );
    c := Coclass(L);
    if c > cc then return false; fi;
    if U.unis and c = cc then return false; fi;

    # a brutal cross-test
    # if Size(Centre(L)) <> Size(Centre(G)) then Error("wrong center"); fi;

    # record that we considered U
    U.ccls := c;
    U.satt := true;
    return true;
end;

#############################################################################
##
#F DiscardSupergroups( G, grps, U )
##
DiscardSupergroups := function( G, grps, U )
    local i, H;
    for i in [1..Length(grps)] do
        if not IsBool(grps[i]) and Length(grps[i].repr) > Length(U.repr) then
            H := SubgroupByIgs(G, grps[i].repr);
            if ForAll( U.repr, x -> x in H ) then
                grps[i] := false;
                Print("        discard group ",i,"\n");
            fi;
        fi;
    od;
end;

#############################################################################
##
#F IsRelevant( G, grps, g, cc, d, deps )
##
IsRelevant := function( G, grps, g, cc, d, deps )
    local i, len;

    # look at the small groups first
    len := List( grps, x -> Length(x.repr) );
    SortParallel( len, grps );

    # filter with centre
    for i in [1..Length(grps)] do
        if not IsBound(grps[i].cent) then
            Print("      check centre in group ",i," of ",Length(grps),"\n");
            if ForAll( grps[i].repr, x -> x*g = g*x ) then 
                Print("        discard group ",i,"\n");
                grps[i] := false; 
            else
                grps[i].cent := true;
            fi;
        fi;
    od;

    if cc = false then return Filtered( grps, x -> x <> false ); fi;

    # filter with coclass
    for i in [1..Length(grps)] do
        if not grps[i] = false and not IsBound(grps[i].satt) and 
           IsSubset(grps[i].rump, deps) 
        then 
            Print("      check coclass in group ",i," of ",Length(grps),"\n");
            if not CheckCoclass( G, grps[i], cc, d, deps ) then 
                DiscardSupergroups( G, grps, grps[i] );
                Print("        discard group ",i,"\n");
                grps[i] := false;
            fi;
        fi;
    od;

    return Filtered( grps, x -> x <> false );
end;

#############################################################################
##
#F SubgroupClassesPGroupByPcp( N, G, D, pcps, max, g, cc, t )
##
SubgroupClassesPGroupByPcp := function(N, G, D, pcps, max, g, cc, t)
    local grps, r, p, s, q, deps, i, new, j, tmp;

    # consider the first step - also to restrict to covered groups
    grps := SubgroupClassesFirstStep( N, G, D, max );

    # set up for iteration
    r := Length(pcps);
    p := N!.prime;
    s := N!.param;
    q := p^(s-1);

    # find the leading deps
    deps := List( pcps, x -> Depth( GeneratorsOfPcp( x )[1] ) );

    # loop over pcps
    for i in [t..r] do
        new := [];
        for j in [1..Length(grps)] do
            if not (deps[i] in grps[j].rump) then
                Print("    layer ",i," of ",r);
                Print(" -- group ",j," of ", Length(grps),"\n");
                tmp := ConsiderClass( N, grps[j], pcps[i], N!.prime );
                Append( new, tmp );
            fi;
        od;
        Append( grps, new );
        grps := IsRelevant( G, grps, g, cc, q, deps{[i+1..r]} );
    od;

    # add a special case
    if r < t then 
        grps := IsRelevant( G, grps, g, cc, q, deps{[t..r]} );
    fi;

    return grps;
end;

#############################################################################
##
#F SubgroupClassesPGroup( N, G, max, g, cc )
##
SubgroupClassesPGroup := function( N, G, max, g, cc )
    local D, hom, fac, efa, pcp, NN, is, mx, GG, gg, grps, i, tmp, DD, t;

    # get series
    Print("    compute series \n");
    D := PRump(G, N!.prime);
    hom := NaturalHomomorphism(N,G);
    fac := List(EfaSeries(Image(hom)), x -> PreImage(hom,x));
    efa := Concatenation( fac, RefinedCentralSeries(N,D) );
    t := Length(fac)+1;

    # translate
    Print("    translate group \n");
    pcp := PcpsBySeries(efa);
    NN := PcpGroupByEfaPcps( pcp );

    Print("    adjust infos \n");
    is := NN!.bijection;
    GG := PreImage( is, G );
    DD := PreImage( is, D );
    gg := PreImage( is, g );
    mx := List( max, x -> PreImage( is, x ) );
  
    # add infos
    NN!.prime := N!.prime;
    NN!.param := N!.param;

    efa := List( efa, x -> PreImage(is, x) );
    pcp := PcpsBySeries( efa );

    # iterate
    Print("    start iteration \n");
    grps := SubgroupClassesPGroupByPcp( NN, GG, DD, pcp, mx, gg, cc, t);

    # translate back and return
    Print("    translate back \n");
    for i in [1..Length(grps)] do
        tmp := rec();

        # the representative
        tmp.repr :=  List( grps[i].repr, x -> Image(is, x) );
        tmp.repr := Cgs(Igs(tmp.repr));

        # the normalizer
        tmp.norm := List( grps[i].norm, x -> Image(is, x) );
        tmp.norm := Cgs(Igs(tmp.norm));

        # add parameters
        tmp.param := G!.param;
        tmp.index := G!.index;
        if IsBound(grps[i].ccls) then tmp.cocla := grps[i].ccls; fi;
        grps[i] := tmp;
    od;

    return grps;
end;

#############################################################################
##
#F UniserialSubgroupsByCentre( Ni, cc ) 
##
UniserialSubgroupsByCentre := function( Ni, cc )
    local N, G, Gi, s, p, i, m, mi, gi, hi, grps;

    N := Parent(Ni);
    G := N!.pointgroup;
    Gi := Ni!.pointgroup;

    s := Ni!.param;
    p := Ni!.prime;
    i := Ni!.index;
    m := G!.maxi;

    if i = 0 and (IsBool(cc) or Coclass(Gi) <= cc) and not COVEREDGP then
        return [rec( repr := Cgs(Gi), norm := Cgs(Ni), index := 0 )];
    elif i = 0 then
        return [];
    fi;

    hi := Product(G!.mins);
    gi := hi^(p^(i-1));
    mi := List( m, x -> NormalIntersection( x, Gi ) );
    if ForAny(mi, x -> Index(Gi,x) <> p) then Error("wrong maxis"); fi;
    
    return SubgroupClassesPGroup( Ni, Gi, mi, gi, cc );
end;

#############################################################################
##
#F UniserialSubgroups( N, cc ) where cc is the desired coclass or false
##
InstallMethod( UniserialSubgroups, true, [IsGroup, IsObject], 0,
function( N, cc )
    local s, p, g, i, Ni;

    # set up
    s := N!.param;
    p := N!.prime;

    # loop
    g := [];
    for i in [0..s-1] do
        Ni := WreathFiniteNormalizer( N, i );
        Append(g, UniserialSubgroupsByCentre( Ni, cc ) );
    od;
    return g;
end );

InstallOtherMethod( UniserialSubgroups, true, [IsGroup], 0,
function(N) return UniserialSubgroups(N,false); end );
        

