#############################################################################
##
#W propgrp.gi                  Coclass package                   Bettina Eick
##
## Some general functions for computing with pro-p-groups by coclass.
##

#############################################################################
##
#A HasFiniteCoclass(G)
##
InstallImmediateMethod( HasFiniteCoclass, IsPcpGroup and IsFinite, 0, 
function(G) return true; end);

InstallMethod( HasFiniteCoclass, true, [IsPcpGroup], 0, 
function(G)
    local N, H, s, A, a, d, p, q, U, V, i;

    if G <> Parent(G) and HasFiniteCoclass(Parent(G)) then return true; fi;

    # get rid of torsion
    N := NormalTorsionSubgroup(G);
    if not IsFinite(N) and IsPGroup(N) then return false; fi;
    H := G/N;

    # get candidates for point group and translations
    s := DerivedSeries(H);
    A := s[Length(s)-1];
    a := Index(H,A);
    d := Length(Pcp(A,"snf"));

    # check a
    if a = infinity then return false; fi; 
    p := Factors(a);
    if Length(Set(p)) > 1 then return false; fi;
    p := p[1];

    # check d
    q := d/(p-1);
    if not IsInt(q) then return false; fi;
    s := Factors(q);
    if Length(Set(s)) > 1 then return false; fi;
    s := Length(s);

    # check for uniserial action
    V := ShallowCopy(A);
    U := ShallowCopy(V);
    for i in [1..d] do
        V := CommutatorSubgroup(U,H);
        if Index(U,V) <> p then return false; fi;
        U := ShallowCopy(V);
    od;

    # This implies that G has finite coclass. Add info.
    SetParameter(G,s);
    SetDimensionProPGroup(G, d);
    SetPrimeProPGroup(G,p);
    SetIsSpaceGroup(G, (Size(N)=1));
   
    # that's it
    return true;
end);

#############################################################################
##
#A PrimeProPGroup(G)
##
InstallMethod( PrimeProPGroup, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    return Factors(RelativeOrdersOfPcp(Pcp(G))[1])[1];
end);
 
#############################################################################
##
#A DimensionProPGroup(G)
##
InstallMethod( DimensionProPGroup,true,[IsPcpGroup and HasFiniteCoclass],0,
function(G)
    return Length(Filtered(RelativeOrdersOfPcp(Pcp(G)), x -> x = 0));
end);

#############################################################################
##
#A Parameter(G)
##
## d = p^{s-1}(p-1) for the parameter s.
##
InstallMethod( Parameter, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local d, p;
    d := DimensionProPGroup(G);
    p := PrimeProPGroup(G);
    return LogInt(d/(p-1),p)+1;
end);

#############################################################################
##
#A HyperCenter(G)
##
InstallMethod( HyperCenter, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    if HasIsSpaceGroup(G) and IsSpaceGroup(G) then
        return TrivialSubgroup(G); 
    fi;
    return NormalTorsionSubgroup(G);
end );

#############################################################################
##
#A TranslationsPre(G)
##
## The preimage of the translations of G/Z_infty(G)
##
InstallMethod( TranslationsPre, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local N, hom, F;
    if IsFinite(G) then return G; fi;
    N := NormalTorsionSubgroup(G);
    hom := NaturalHomomorphism(G,N);
    F := FittingSubgroup(Image(hom));
    return PreImage(hom, F);
end );

#############################################################################
##
#A TranslationsLow(G)
##
## Returns gamma_i(G) with i minimal so that gamma_i(G) is free abelian.
##
InstallMethod( TranslationsLow, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local T;
    if IsFinite(G) then return TrivialSubgroup(G); fi;
    T := G;
    while not IsFreeAbelian(T) do
        T := CommutatorSubgroup(T,G);
        T := SubgroupByIgs(G, Cgs(T));
    od;
    return T;
end);

#############################################################################
##
#A Translations(G)
##
## Returns some translation subgroup of G.
##
InstallMethod( Translations, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local T;
    
    # first try the obvious
    T := SubgroupByIgs(G, Filtered(Igs(G), x -> RelativeOrderPcp(x)=0));
    if IsNormal(G,T) and IsFreeAbelian(T) then return T; fi;
   
    # otherwise 
    return TranslationsLow(G);
end );

#############################################################################
##
#A PointGroupPre(G)
##
InstallMethod( PointGroupPre, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    return G/TranslationsPre(G);
end );

#############################################################################
##
#A PointGroupLow(G)
##
InstallMethod( PointGroupLow, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    return G/TranslationsLow(G);
end );

#############################################################################
##
#A PointGroup(G)
##
InstallMethod( PointGroup, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    return G/Translations(G);
end );

#############################################################################
##
#A Coclass(G)
##
InstallMethod( Coclass, true, [IsPcpGroup], 0,
function(G)
    local H, c, n;
    if IsFinite(G) then 
        H := G;
    elif HasFiniteCoclass(G) then
        H := PointGroupLow(G);
    else
        return fail; 
    fi;
    c := Length(LowerCentralSeries(H)) - 1;
    n := Length(Factors(Size(H)));
    return n-c;
end );

#############################################################################
##
#A IsCFGroup(G) and IsECFGroup(G)
##
InstallMethod( IsCFGroup, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local l, f;
    l := LowerCentralSeries(PointGroupLow(G));
    f := List([2..Length(l)-1], x -> Index(l[x], l[x+1]));
    return ForAll(f, x -> IsPrime(x));
end );

InstallMethod( IsECFGroup, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local l, f;
    l := LowerCentralSeries(PointGroupLow(G));
    f := List([2..Length(l)-1], x -> Index(l[x], l[x+1]));
    if not ForAll(f, x -> IsPrime(x)) then return false; fi;
    return IsElementaryAbelian(l[1]/l[2]);
end );

#############################################################################
##
#A CentralExponent(G)
##
InstallMethod( CentralExponent, true, [IsPcpGroup and HasFiniteCoclass], 0,
function(G)
    local C;
    C := Center(PointGroupPre(G));
    return Length(Factors(Size(C)));
end );




